@@ -12,11 +12,7 @@ where
     T: PolarsDataType,
 {
     if N::get_dtype() == T::get_dtype() {
-        // convince the compiler that N and T are the same type
-        return unsafe {
-            let ca = std::mem::transmute(ca.clone());
-            Ok(ca)
-        };
+        return Ok(ChunkedArray::new_from_chunks(ca.name(), ca.chunks.clone()));
     };
     let chunks = ca
         .chunks
@@ -79,7 +75,12 @@ macro_rules! cast_with_dtype {
             List(_) => ChunkCast::cast::<ListType>($self).map(|ca| ca.into_series()),
             Categorical => ChunkCast::cast::<CategoricalType>($self).map(|ca| ca.into_series()),
             dt => Err(PolarsError::Other(
-                format!("Casting to {:?} is not supported", dt).into(),
+                format!(
+                    "Casting to {:?} is not supported. \
+                This error may occur because you did not activate a certain dtype feature",
+                    dt
+                )
+                .into(),
             )),
         }
     }};
@@ -116,6 +117,11 @@ impl ChunkCast for CategoricalChunked {
                 ca.field = Arc::new(Field::new(ca.name(), DataType::UInt32));
                 Ok(ca)
             }
+            DataType::Categorical => {
+                let mut out = ChunkedArray::new_from_chunks(self.name(), self.chunks.clone());
+                out.categorical_map = self.categorical_map.clone();
+                Ok(out)
+            }
             _ => cast_ca(self),
         }
     }
@@ -199,15 +205,39 @@ impl ChunkCast for Utf8Chunked {
     }
 }
 
+fn boolean_to_utf8(ca: &BooleanChunked) -> Utf8Chunked {
+    ca.into_iter()
+        .map(|opt_b| match opt_b {
+            Some(true) => Some("true"),
+            Some(false) => Some("false"),
+            None => None,
+        })
+        .collect()
+}
+
 impl ChunkCast for BooleanChunked {
     fn cast<N>(&self) -> Result<ChunkedArray<N>>
     where
         N: PolarsDataType,
     {
-        cast_ca(self)
+        if matches!(N::get_dtype(), DataType::Utf8) {
+            let mut ca = boolean_to_utf8(self);
+            Ok(ChunkedArray::new_from_chunks(
+                self.name(),
+                std::mem::take(&mut ca.chunks),
+            ))
+        } else {
+            cast_ca(self)
+        }
     }
     fn cast_with_dtype(&self, data_type: &DataType) -> Result<Series> {
-        cast_with_dtype!(self, data_type)
+        if matches!(data_type, DataType::Utf8) {
+            let mut ca = boolean_to_utf8(self);
+            ca.rename(self.name());
+            Ok(ca.into_series())
+        } else {
+            cast_with_dtype!(self, data_type)
+        }
     }
 }
 
@@ -239,9 +269,11 @@ impl ChunkCast for ListChunked {
                         Ok(make_array(new))
                     })
                     .collect::<Result<_>>()?;
-                let ca = ListChunked::new_from_chunks(self.name(), chunks);
-                let ca = unsafe { std::mem::transmute(ca) };
-                Ok(ca)
+                let mut ca = ListChunked::new_from_chunks(self.name(), chunks);
+                Ok(ChunkedArray::new_from_chunks(
+                    self.name(),
+                    std::mem::take(&mut ca.chunks),
+                ))
             }
             _ => Err(PolarsError::Other("Cannot cast list type".into())),
         }
@@ -293,4 +325,13 @@ mod test {
         assert_eq!(new.dtype(), &DataType::List(ArrowDataType::Float64));
         Ok(())
     }
+
+    #[test]
+    fn test_cast_noop() {
+        // check if we can cast categorical twice without panic
+        let ca = Utf8Chunked::new_from_slice("foo", &["bar", "ham"]);
+        let out = ca.cast::<CategoricalType>().unwrap();
+        let out = out.cast::<CategoricalType>().unwrap();
+        assert_eq!(out.dtype(), &DataType::Categorical)
+    }
 }
