@@ -178,7 +178,7 @@ impl<T> ChunkedArray<T> {
                     return Some(offset);
                 } else {
                     let arr = &self.chunks[idx];
-                    let null_bit_buffer = null_bit_buffer.as_ref().unwrap();
+                    let null_bit_buffer = null_bit_buffer.unwrap();
                     let bit_end = arr.offset() + arr.len();
 
                     let byte_start = std::cmp::min(round_upto_power_of_2(arr.offset(), 8), bit_end);
@@ -196,17 +196,25 @@ impl<T> ChunkedArray<T> {
         }
     }
 
-    /// Get the null count and the buffer of bits representing null values
-    pub fn null_bits(&self) -> impl Iterator<Item = (usize, Option<Buffer>)> + '_ {
+    /// Get the null count and the buffer of bits representing the validity values
+    pub fn null_bits(&self) -> impl Iterator<Item = (usize, Option<&Buffer>)> + '_ {
         self.chunks.iter().map(|arr| get_bitmap(arr.as_ref()))
     }
 
+    /// Shrink the capacity of this array to fit it's length.
+    pub fn shrink_to_fit(&mut self) {
+        self.chunks = vec![arrow::compute::concat(
+            &self.chunks.iter().map(|a| &**a).collect_vec().as_slice(),
+        )
+        .unwrap()];
+    }
+
     /// Unpack a Series to the same physical type.
     ///
     /// # Safety
     ///
     /// This is unsafe as the dtype may be uncorrect and
-    /// is assumed to be correct in other unsafe code.
+    /// is assumed to be correct in other safe code.
     pub(crate) unsafe fn unpack_series_matching_physical_type(
         &self,
         series: &Series,
@@ -1020,4 +1028,23 @@ pub(crate) mod test {
         let v: Vec<_> = ca.into_iter().collect();
         assert_eq!(v, &[Some(0), None, Some(1), Some(2)]);
     }
+
+    #[test]
+    fn test_shrink_to_fit() {
+        let mut builder = Utf8ChunkedBuilder::new("foo", 2048, 100 * 2048);
+        builder.append_value("foo");
+        let mut arr = builder.finish();
+        let before = arr
+            .chunks()
+            .iter()
+            .map(|arr| arr.get_buffer_memory_size())
+            .sum::<usize>();
+        arr.shrink_to_fit();
+        let after = arr
+            .chunks()
+            .iter()
+            .map(|arr| arr.get_buffer_memory_size())
+            .sum::<usize>();
+        assert!(before > after);
+    }
 }
