@@ -1,31 +1,25 @@
+use crate::chunked_array::builder::categorical::RevMapping;
 #[cfg(feature = "object")]
 use crate::chunked_array::object::ObjectType;
-use crate::frame::groupby::GroupTuples;
+use crate::datatypes::PlHashSet;
+use crate::frame::groupby::{GroupTuples, IntoGroupTuples};
 use crate::prelude::*;
-use crate::utils::{floating_encode_f64, integer_decode_f64, NoNull};
-use crate::{chunked_array::float::IntegerDecode, frame::groupby::IntoGroupTuples};
-use ahash::RandomState;
+use crate::utils::NoNull;
+use arrow::array::Array;
 use itertools::Itertools;
-use num::{NumCast, ToPrimitive};
+use num::NumCast;
 use rayon::prelude::*;
-use std::collections::HashSet;
 use std::fmt::Display;
 use std::hash::Hash;
 
-pub(crate) fn is_unique_helper(
-    mut groups: GroupTuples,
+fn finish_is_unique_helper(
+    mut unique_idx: Vec<u32>,
     len: u32,
     unique_val: bool,
     duplicated_val: bool,
 ) -> BooleanChunked {
-    debug_assert_ne!(unique_val, duplicated_val);
-    groups.sort_unstable_by_key(|t| t.0);
-
-    let mut unique_idx_iter = groups
-        .into_iter()
-        .filter(|(_, g)| g.len() == 1)
-        .map(|(first, _)| first);
-
+    unique_idx.sort_unstable();
+    let mut unique_idx_iter = unique_idx.into_iter();
     let mut next_unique_idx = unique_idx_iter.next();
     (0..len)
         .into_iter()
@@ -43,26 +37,53 @@ pub(crate) fn is_unique_helper(
         .collect()
 }
 
-fn is_unique<T>(ca: &ChunkedArray<T>) -> BooleanChunked
-where
-    T: PolarsDataType,
-    ChunkedArray<T>: IntoGroupTuples,
-{
-    let groups = ca.group_tuples(true);
-    let mut out = is_unique_helper(groups, ca.len() as u32, true, false);
-    out.rename(ca.name());
-    out
+pub(crate) fn is_unique_helper2(
+    unique_idx: Vec<u32>,
+    len: u32,
+    unique_val: bool,
+    duplicated_val: bool,
+) -> BooleanChunked {
+    debug_assert_ne!(unique_val, duplicated_val);
+    finish_is_unique_helper(unique_idx, len, unique_val, duplicated_val)
 }
 
-fn is_duplicated<T>(ca: &ChunkedArray<T>) -> BooleanChunked
-where
-    T: PolarsDataType,
-    ChunkedArray<T>: IntoGroupTuples,
-{
-    let groups = ca.group_tuples(true);
-    let mut out = is_unique_helper(groups, ca.len() as u32, false, true);
-    out.rename(ca.name());
-    out
+pub(crate) fn is_unique_helper(
+    groups: GroupTuples,
+    len: u32,
+    unique_val: bool,
+    duplicated_val: bool,
+) -> BooleanChunked {
+    debug_assert_ne!(unique_val, duplicated_val);
+    let idx = groups
+        .into_iter()
+        .filter_map(|(first, g)| if g.len() == 1 { Some(first) } else { None })
+        .collect::<Vec<_>>();
+    finish_is_unique_helper(idx, len, unique_val, duplicated_val)
+}
+
+/// if inverse is true, this is an `is_duplicated`
+/// otherwise an `is_unique`
+macro_rules! is_unique_duplicated {
+    ($ca:expr, $inverse:expr) => {{
+        let mut idx_key = PlHashMap::new();
+
+        // instead of grouptuples, wich allocates a full vec per group, we now just toggle a boolean
+        // that's false if a group has multiple entries.
+        $ca.into_iter().enumerate().for_each(|(idx, key)| {
+            idx_key
+                .entry(key)
+                .and_modify(|v: &mut (u32, bool)| v.1 = false)
+                .or_insert((idx as u32, true));
+        });
+
+        let idx: Vec<_> = idx_key
+            .into_iter()
+            .filter_map(|(_k, v)| if v.1 { Some(v.0) } else { None })
+            .collect();
+        let mut out = is_unique_helper2(idx, $ca.len() as u32, !$inverse, $inverse);
+        out.rename($ca.name());
+        Ok(out)
+    }};
 }
 
 impl ChunkUnique<ListType> for ListChunked {
@@ -93,24 +114,18 @@ impl<T> ChunkUnique<ObjectType<T>> for ObjectChunked<T> {
     }
 }
 
-fn fill_set<A>(a: impl Iterator<Item = A>) -> HashSet<A, RandomState>
+fn fill_set<A>(a: impl Iterator<Item = A>) -> PlHashSet<A>
 where
     A: Hash + Eq,
 {
-    let mut set = HashSet::with_hasher(RandomState::new());
-
-    for val in a {
-        set.insert(val);
-    }
-
-    set
+    a.collect()
 }
 
 fn arg_unique<T>(a: impl Iterator<Item = T>, capacity: usize) -> AlignedVec<u32>
 where
     T: Hash + Eq,
 {
-    let mut set = HashSet::with_hasher(RandomState::new());
+    let mut set = PlHashSet::new();
     let mut unique = AlignedVec::with_capacity_aligned(capacity);
     a.enumerate().for_each(|(idx, val)| {
         if set.insert(val) {
@@ -164,16 +179,20 @@ where
     }
 
     fn is_unique(&self) -> Result<BooleanChunked> {
-        Ok(is_unique(self))
+        is_unique_duplicated!(self, false)
     }
 
     fn is_duplicated(&self) -> Result<BooleanChunked> {
-        Ok(is_duplicated(self))
+        is_unique_duplicated!(self, true)
     }
 
     fn value_counts(&self) -> Result<DataFrame> {
         impl_value_counts!(self)
     }
+
+    fn n_unique(&self) -> Result<usize> {
+        Ok(fill_set(self.into_iter()).len())
+    }
 }
 
 impl ChunkUnique<Utf8Type> for Utf8Chunked {
@@ -193,53 +212,58 @@ impl ChunkUnique<Utf8Type> for Utf8Chunked {
     }
 
     fn is_unique(&self) -> Result<BooleanChunked> {
-        Ok(is_unique(self))
+        is_unique_duplicated!(self, false)
     }
     fn is_duplicated(&self) -> Result<BooleanChunked> {
-        Ok(is_duplicated(self))
+        is_unique_duplicated!(self, true)
     }
 
     fn value_counts(&self) -> Result<DataFrame> {
         impl_value_counts!(self)
     }
+
+    fn n_unique(&self) -> Result<usize> {
+        Ok(fill_set(self.into_iter()).len())
+    }
 }
 
 impl ChunkUnique<CategoricalType> for CategoricalChunked {
     fn unique(&self) -> Result<Self> {
-        let set = fill_set(self.into_iter());
-        let mut ca = UInt32Chunked::new_from_opt_iter(self.name(), set.iter().copied());
+        let cat_map = self.categorical_map.as_ref().unwrap();
+        let mut ca = match &**cat_map {
+            RevMapping::Local(a) => UInt32Chunked::new_from_iter(self.name(), 0..(a.len() as u32)),
+            RevMapping::Global(map, _, _) => {
+                UInt32Chunked::new_from_iter(self.name(), map.keys().copied())
+            }
+        };
         ca.categorical_map = self.categorical_map.clone();
         ca.cast()
     }
 
     fn arg_unique(&self) -> Result<UInt32Chunked> {
-        Ok(UInt32Chunked::new_from_aligned_vec(
-            self.name(),
-            arg_unique_ca!(self),
-        ))
+        self.cast::<UInt32Type>()?.arg_unique()
     }
 
     fn is_unique(&self) -> Result<BooleanChunked> {
-        Ok(is_unique(self))
+        self.cast::<UInt32Type>()?.is_unique()
     }
     fn is_duplicated(&self) -> Result<BooleanChunked> {
-        Ok(is_duplicated(self))
+        self.cast::<UInt32Type>()?.is_duplicated()
     }
 
     fn value_counts(&self) -> Result<DataFrame> {
         impl_value_counts!(self)
     }
+    fn n_unique(&self) -> Result<usize> {
+        Ok(self.categorical_map.as_ref().unwrap().len())
+    }
 }
 
 #[cfg(feature = "dtype-u8")]
 fn dummies_helper(mut groups: Vec<u32>, len: usize, name: &str) -> UInt8Chunked {
     groups.sort_unstable();
 
-    // let mut group_member_iter = groups.into_iter();
-    let mut av = AlignedVec::with_capacity_aligned(len);
-    for _ in 0..len {
-        av.push(0u8)
-    }
+    let mut av: AlignedVec<_> = (0..len).map(|_| 0u8).collect();
 
     for idx in groups {
         let elem = unsafe { av.inner.get_unchecked_mut(idx as usize) };
@@ -250,14 +274,11 @@ fn dummies_helper(mut groups: Vec<u32>, len: usize, name: &str) -> UInt8Chunked
 }
 
 #[cfg(not(feature = "dtype-u8"))]
-fn dummies_helper(mut groups: Vec<u32>, len: usize, name: &str) -> Int64Chunked {
+fn dummies_helper(mut groups: Vec<u32>, len: usize, name: &str) -> Int32Chunked {
     groups.sort_unstable();
 
     // let mut group_member_iter = groups.into_iter();
-    let mut av = AlignedVec::with_capacity_aligned(len);
-    for _ in 0..len {
-        av.push(0i64)
-    }
+    let mut av: AlignedVec<_> = (0..len).map(|_| 0i32).collect();
 
     for idx in groups {
         let elem = unsafe { av.inner.get_unchecked_mut(idx as usize) };
@@ -347,73 +368,32 @@ impl ChunkUnique<BooleanType> for BooleanChunked {
     }
 
     fn is_unique(&self) -> Result<BooleanChunked> {
-        Ok(is_unique(self))
+        is_unique_duplicated!(self, false)
     }
     fn is_duplicated(&self) -> Result<BooleanChunked> {
-        Ok(is_duplicated(self))
-    }
-}
-
-fn float_unique<T>(ca: &ChunkedArray<T>) -> ChunkedArray<T>
-where
-    T: PolarsFloatType,
-    T::Native: NumCast + ToPrimitive,
-{
-    let set = match ca.null_count() {
-        0 => fill_set(
-            ca.into_no_null_iter()
-                .map(|v| Some(integer_decode_f64(v.to_f64().unwrap()))),
-        ),
-        _ => fill_set(
-            ca.into_iter()
-                .map(|opt_v| opt_v.map(|v| integer_decode_f64(v.to_f64().unwrap()))),
-        ),
-    };
-    ChunkedArray::new_from_opt_iter(
-        ca.name(),
-        set.iter().copied().map(|opt| match opt {
-            Some((mantissa, exponent, sign)) => {
-                let flt = floating_encode_f64(mantissa, exponent, sign);
-                let val: T::Native = NumCast::from(flt).unwrap();
-                Some(val)
-            }
-            None => None,
-        }),
-    )
-}
-
-fn float_arg_unique<T>(ca: &ChunkedArray<T>) -> AlignedVec<u32>
-where
-    T: PolarsFloatType,
-    T::Native: IntegerDecode,
-{
-    match ca.null_count() {
-        0 => arg_unique(ca.into_no_null_iter().map(|v| v.integer_decode()), ca.len()),
-        _ => arg_unique(
-            ca.into_iter()
-                .map(|opt_v| opt_v.map(|v| v.integer_decode())),
-            ca.len(),
-        ),
+        is_unique_duplicated!(self, true)
     }
 }
 
 impl ChunkUnique<Float32Type> for Float32Chunked {
     fn unique(&self) -> Result<ChunkedArray<Float32Type>> {
-        Ok(float_unique(self))
+        let ca = self.bit_repr_small();
+        let set = fill_set(ca.into_iter());
+        Ok(set
+            .into_iter()
+            .map(|opt_v| opt_v.map(f32::from_bits))
+            .collect())
     }
 
     fn arg_unique(&self) -> Result<UInt32Chunked> {
-        Ok(UInt32Chunked::new_from_aligned_vec(
-            self.name(),
-            float_arg_unique(self),
-        ))
+        self.bit_repr_small().arg_unique()
     }
 
     fn is_unique(&self) -> Result<BooleanChunked> {
-        Ok(is_unique(self))
+        self.bit_repr_small().is_unique()
     }
     fn is_duplicated(&self) -> Result<BooleanChunked> {
-        Ok(is_duplicated(self))
+        self.bit_repr_small().is_duplicated()
     }
     fn value_counts(&self) -> Result<DataFrame> {
         impl_value_counts!(self)
@@ -422,21 +402,47 @@ impl ChunkUnique<Float32Type> for Float32Chunked {
 
 impl ChunkUnique<Float64Type> for Float64Chunked {
     fn unique(&self) -> Result<ChunkedArray<Float64Type>> {
-        Ok(float_unique(self))
+        let ca = self.bit_repr_large();
+        let set = fill_set(ca.into_iter());
+        Ok(set
+            .into_iter()
+            .map(|opt_v| opt_v.map(f64::from_bits))
+            .collect())
     }
 
     fn arg_unique(&self) -> Result<UInt32Chunked> {
-        Ok(UInt32Chunked::new_from_aligned_vec(
-            self.name(),
-            float_arg_unique(self),
-        ))
+        #[cfg(feature = "dtype-u64")]
+        {
+            self.bit_repr_large().arg_unique()
+        }
+
+        #[cfg(not(feature = "dtype-u64"))]
+        {
+            panic!("activate feature dtype-u64")
+        }
     }
 
     fn is_unique(&self) -> Result<BooleanChunked> {
-        Ok(is_unique(self))
+        #[cfg(feature = "dtype-u64")]
+        {
+            self.bit_repr_large().is_unique()
+        }
+
+        #[cfg(not(feature = "dtype-u64"))]
+        {
+            panic!("activate feature dtype-u64")
+        }
     }
     fn is_duplicated(&self) -> Result<BooleanChunked> {
-        Ok(is_duplicated(self))
+        #[cfg(feature = "dtype-u64")]
+        {
+            self.bit_repr_large().is_duplicated()
+        }
+
+        #[cfg(not(feature = "dtype-u64"))]
+        {
+            panic!("activate feature dtype-u64")
+        }
     }
     fn value_counts(&self) -> Result<DataFrame> {
         impl_value_counts!(self)
