@@ -2,7 +2,7 @@ use self::hashing::*;
 use crate::chunked_array::builder::PrimitiveChunkedBuilder;
 use crate::frame::select::Selection;
 use crate::prelude::*;
-use crate::utils::{accumulate_dataframes_vertical, split_ca, NoNull};
+use crate::utils::{accumulate_dataframes_vertical, set_partition_size, split_ca, NoNull};
 use crate::vector_hasher::{AsU64, StrHash};
 use crate::POOL;
 use ahash::RandomState;
@@ -49,43 +49,27 @@ where
         0
     };
     if multithreaded && group_multithreaded(ca) {
-        let n_threads = POOL.current_num_threads();
-        let splitted = split_ca(ca, n_threads).unwrap();
+        let n_partitions = set_partition_size() as u64;
 
         // use the arrays as iterators
         if ca.chunks.len() == 1 {
             if ca.null_count() == 0 {
-                let keys = splitted
-                    .iter()
-                    .map(|ca| ca.cont_slice().unwrap())
-                    .collect::<Vec<_>>();
-                groupby_threaded_num(keys, group_size_hint)
+                let keys = vec![ca.cont_slice().unwrap()];
+                groupby_threaded_num(keys, group_size_hint, n_partitions)
             } else {
-                let keys = splitted
-                    .iter()
-                    .map(|ca| {
-                        ca.downcast_iter()
-                            .map(|v| v.into_iter().map(|v| v.as_u64()))
-                            .flatten()
-                            .collect::<Vec<_>>()
-                    })
+                let keys = ca
+                    .downcast_iter()
+                    .map(|arr| arr.into_iter().collect::<Vec<_>>())
                     .collect::<Vec<_>>();
-
-                groupby_threaded_num(keys, group_size_hint)
+                groupby_threaded_num(keys, group_size_hint, n_partitions)
             }
             // use the polars-iterators
         } else if ca.null_count() == 0 {
-            let keys = splitted
-                .iter()
-                .map(|ca| ca.into_no_null_iter().collect::<Vec<_>>())
-                .collect::<Vec<_>>();
-            groupby_threaded_num(keys, group_size_hint)
+            let keys = vec![ca.into_no_null_iter().collect::<Vec<_>>()];
+            groupby_threaded_num(keys, group_size_hint, n_partitions)
         } else {
-            let keys = splitted
-                .iter()
-                .map(|ca| ca.into_iter().map(|v| v.as_u64()).collect::<Vec<_>>())
-                .collect::<Vec<_>>();
-            groupby_threaded_num(keys, group_size_hint)
+            let keys = vec![ca.into_iter().collect::<Vec<_>>()];
+            groupby_threaded_num(keys, group_size_hint, n_partitions)
         }
     } else if ca.null_count() == 0 {
         groupby(ca.into_no_null_iter())
@@ -142,8 +126,9 @@ impl IntoGroupTuples for Utf8Chunked {
     fn group_tuples(&self, multithreaded: bool) -> GroupTuples {
         let hb = RandomState::default();
         if multithreaded {
-            let n_threads = POOL.current_num_threads();
-            let splitted = split_ca(self, n_threads).unwrap();
+            let n_partitions = set_partition_size();
+
+            let splitted = split_ca(self, n_partitions).unwrap();
 
             let str_hashes = POOL.install(|| {
                 splitted
@@ -160,7 +145,7 @@ impl IntoGroupTuples for Utf8Chunked {
                     })
                     .collect::<Vec<_>>()
             });
-            groupby_threaded_num(str_hashes, 0)
+            groupby_threaded_num(str_hashes, 0, n_partitions as u64)
         } else {
             let str_hashes = self
                 .into_iter()
@@ -188,8 +173,142 @@ impl IntoGroupTuples for ListChunked {}
 #[cfg(feature = "object")]
 impl<T> IntoGroupTuples for ObjectChunked<T> {}
 
+/// Used to tightly two 32 bit values and null information
+/// Only the bit values matter, not the meaning of the bits
+fn pack_u32_tuples(opt_l: Option<u32>, opt_r: Option<u32>) -> [u8; 9] {
+    // 4 bytes for first value
+    // 4 bytes for second value
+    // last bytes' bits are used to indicate missing values
+    let mut val = [0u8; 9];
+    let s = &mut val;
+    match (opt_l, opt_r) {
+        (Some(l), Some(r)) => {
+            // write to first 4 places
+            (&mut s[..4]).copy_from_slice(&l.to_ne_bytes());
+            // write to second chunk of 4 places
+            (&mut s[4..8]).copy_from_slice(&r.to_ne_bytes());
+            // leave last byte as is
+        }
+        (Some(l), None) => {
+            (&mut s[..4]).copy_from_slice(&l.to_ne_bytes());
+            // set right null bit
+            s[8] = 1;
+        }
+        (None, Some(r)) => {
+            (&mut s[4..8]).copy_from_slice(&r.to_ne_bytes());
+            // set left null bit
+            s[8] = 1 << 1;
+        }
+        (None, None) => {
+            // set two null bits
+            s[8] = 3;
+        }
+    }
+    val
+}
+
+/// Used to tightly two 64 bit values and null information
+/// Only the bit values matter, not the meaning of the bits
+fn pack_u64_tuples(opt_l: Option<u64>, opt_r: Option<u64>) -> [u8; 17] {
+    // 8 bytes for first value
+    // 8 bytes for second value
+    // last bytes' bits are used to indicate missing values
+    let mut val = [0u8; 17];
+    let s = &mut val;
+    match (opt_l, opt_r) {
+        (Some(l), Some(r)) => {
+            // write to first 4 places
+            (&mut s[..8]).copy_from_slice(&l.to_ne_bytes());
+            // write to second chunk of 4 places
+            (&mut s[8..16]).copy_from_slice(&r.to_ne_bytes());
+            // leave last byte as is
+        }
+        (Some(l), None) => {
+            (&mut s[..8]).copy_from_slice(&l.to_ne_bytes());
+            // set right null bit
+            s[16] = 1;
+        }
+        (None, Some(r)) => {
+            (&mut s[8..16]).copy_from_slice(&r.to_ne_bytes());
+            // set left null bit
+            s[16] = 1 << 1;
+        }
+        (None, None) => {
+            // set two null bits
+            s[16] = 3;
+        }
+    }
+    val
+}
+
+/// Used to tightly one 32 bit and a 64 bit valued type and null information
+/// Only the bit values matter, not the meaning of the bits
+fn pack_u32_u64_tuples(opt_l: Option<u32>, opt_r: Option<u64>) -> [u8; 13] {
+    // 8 bytes for first value
+    // 8 bytes for second value
+    // last bytes' bits are used to indicate missing values
+    let mut val = [0u8; 13];
+    let s = &mut val;
+    match (opt_l, opt_r) {
+        (Some(l), Some(r)) => {
+            // write to first 4 places
+            (&mut s[..4]).copy_from_slice(&l.to_ne_bytes());
+            // write to second chunk of 4 places
+            (&mut s[4..12]).copy_from_slice(&r.to_ne_bytes());
+            // leave last byte as is
+        }
+        (Some(l), None) => {
+            (&mut s[..4]).copy_from_slice(&l.to_ne_bytes());
+            // set right null bit
+            s[12] = 1;
+        }
+        (None, Some(r)) => {
+            (&mut s[4..12]).copy_from_slice(&r.to_ne_bytes());
+            // set left null bit
+            s[12] = 1 << 1;
+        }
+        (None, None) => {
+            // set two null bits
+            s[12] = 3;
+        }
+    }
+    val
+}
+
 impl DataFrame {
     pub fn groupby_with_series(&self, by: Vec<Series>, multithreaded: bool) -> Result<GroupBy> {
+        macro_rules! finish_packed_bit_path {
+            ($ca0:expr, $ca1:expr, $pack_fn:expr) => {{
+                let n_partitions = set_partition_size();
+
+                // we split so that we can prepare the data over multiple threads.
+                // pack the bit values together and add a final byte that will be 0
+                // when there are no null values.
+                // otherwise we use two bits of this byte to represent null values.
+                let split_0 = split_ca(&$ca0, n_partitions).unwrap();
+                let split_1 = split_ca(&$ca1, n_partitions).unwrap();
+                let keys = POOL.install(|| {
+                    split_0
+                        .into_par_iter()
+                        .zip(split_1.into_par_iter())
+                        .map(|(ca0, ca1)| {
+                            ca0.into_iter()
+                                .zip(ca1.into_iter())
+                                .map(|(l, r)| $pack_fn(l, r))
+                                .collect::<Vec<_>>()
+                        })
+                        .collect::<Vec<_>>()
+                });
+
+                return Ok(GroupBy::new(
+                    self,
+                    by,
+                    groupby_threaded_num(keys, 0, n_partitions as u64),
+                    None,
+                ));
+            }};
+        }
+
         if by.is_empty() || by[0].len() != self.height() {
             return Err(PolarsError::ShapeMisMatch(
                 "the Series used as keys should have the same length as the DataFrame".into(),
@@ -201,7 +320,18 @@ impl DataFrame {
             by.iter()
                 .map(|s| match s.dtype() {
                     DataType::Categorical => s.cast::<UInt32Type>().unwrap(),
-                    _ => s.clone(),
+                    DataType::Float32 => s.bit_repr_small().into_series(),
+                    // otherwise we use the vec hash for float
+                    #[cfg(feature = "dtype-u64")]
+                    DataType::Float64 => s.bit_repr_large().into_series(),
+                    _ => {
+                        // is date like
+                        if !s.is_numeric() && s.is_numeric_physical() {
+                            s.to_physical_repr()
+                        } else {
+                            s.clone()
+                        }
+                    }
                 })
                 .collect(),
         )?;
@@ -212,12 +342,42 @@ impl DataFrame {
                 series.group_tuples(multithreaded)
             }
             _ => {
-                if multithreaded {
-                    let n_threads = num_cpus::get();
-                    groupby_threaded_multiple_keys_flat(keys_df, n_threads)
-                } else {
-                    groupby_multiple_keys(keys_df)
+                // multiple keys is always multi-threaded
+                // reduce code paths
+                let s0 = &keys_df.get_columns()[0];
+                let s1 = &keys_df.get_columns()[1];
+
+                // fast path for numeric data
+                // uses the bit values to tightly pack those into arrays.
+                if by.len() == 2 && s0.is_numeric() && s1.is_numeric() {
+                    match (s0.bit_repr_is_large(), s1.bit_repr_is_large()) {
+                        (false, false) => {
+                            let ca0 = s0.bit_repr_small();
+                            let ca1 = s1.bit_repr_small();
+                            finish_packed_bit_path!(ca0, ca1, pack_u32_tuples)
+                        }
+                        (true, true) => {
+                            let ca0 = s0.bit_repr_large();
+                            let ca1 = s1.bit_repr_large();
+                            finish_packed_bit_path!(ca0, ca1, pack_u64_tuples)
+                        }
+                        (true, false) => {
+                            let ca0 = s0.bit_repr_large();
+                            let ca1 = s1.bit_repr_small();
+                            // small first
+                            finish_packed_bit_path!(ca1, ca0, pack_u32_u64_tuples)
+                        }
+                        (false, true) => {
+                            let ca0 = s0.bit_repr_small();
+                            let ca1 = s1.bit_repr_large();
+                            // small first
+                            finish_packed_bit_path!(ca0, ca1, pack_u32_u64_tuples)
+                        }
+                    }
                 }
+
+                let n_partitions = set_partition_size();
+                groupby_threaded_multiple_keys_flat(keys_df, n_partitions)
             }
         };
         Ok(GroupBy::new(self, by, groups, None))
@@ -1295,7 +1455,7 @@ mod test {
             let a = groupby(ca.into_iter()).into_iter().sorted().collect_vec();
 
             let keys = splitted.iter().map(|ca| ca.cont_slice().unwrap()).collect();
-            let b = groupby_threaded_num(keys, 0)
+            let b = groupby_threaded_num(keys, 0, splitted.len() as u64)
                 .into_iter()
                 .sorted()
                 .collect_vec();
@@ -1344,4 +1504,20 @@ mod test {
         assert!((val - expected).abs() < 0.000001);
         Ok(())
     }
+
+    #[test]
+    fn test_groupby_null_group() -> Result<()> {
+        // check if null is own group
+        let mut df = df![
+            "g" => [Some("foo"), Some("foo"), Some("bar"), None, None],
+            "flt" => [1.0, 2.0, 3.0, 1.0, 1.0],
+            "int" => [1, 2, 3, 1, 1]
+        ]?;
+
+        df.may_apply("g", |s| s.cast::<CategoricalType>())?;
+
+        let out = df.groupby("g")?.sum()?;
+        dbg!(out);
+        Ok(())
+    }
 }
