@@ -5,12 +5,12 @@ use crate::frame::hash_join::multiple_keys::{
 };
 use crate::frame::select::Selection;
 use crate::prelude::*;
-use crate::utils::{split_ca, NoNull};
+use crate::utils::{set_partition_size, split_ca, NoNull};
 use crate::vector_hasher::{
-    create_hash_and_keys_threaded_vectorized, prepare_hashed_relation_threaded, this_thread, AsU64,
-    StrHash,
+    create_hash_and_keys_threaded_vectorized, prepare_hashed_relation_threaded, this_partition,
+    AsU64, StrHash,
 };
-use crate::POOL;
+use crate::{datatypes::PlHashMap, POOL};
 use ahash::RandomState;
 use hashbrown::hash_map::{Entry, RawEntryMut};
 use hashbrown::HashMap;
@@ -64,28 +64,32 @@ pub enum JoinType {
     Outer,
 }
 
-unsafe fn get_hash_tbl_threaded_join<T, H>(
+unsafe fn get_hash_tbl_threaded_join_partitioned<T, H>(
     h: u64,
     hash_tables: &[HashMap<T, Vec<u32>, H>],
     len: u64,
 ) -> &HashMap<T, Vec<u32>, H> {
     let mut idx = 0;
     for i in 0..len {
-        if (h + i) % len == 0 {
+        // can only be done for powers of two.
+        // n % 2^i = n & (2^i - 1)
+        if (h + i) & (len - 1) == 0 {
             idx = i as usize;
         }
     }
     hash_tables.get_unchecked(idx)
 }
 
-unsafe fn get_hash_tbl_threaded_join_mut<T, H>(
+unsafe fn get_hash_tbl_threaded_join_mut_partitioned<T, H>(
     h: u64,
     hash_tables: &mut [HashMap<T, Vec<u32>, H>],
     len: u64,
 ) -> &mut HashMap<T, Vec<u32>, H> {
     let mut idx = 0;
     for i in 0..len {
-        if (h + i) % len == 0 {
+        // can only be done for powers of two.
+        // n % 2^i = n & (2^i - 1)
+        if (h + i) & (len - 1) == 0 {
             idx = i as usize;
         }
     }
@@ -95,7 +99,7 @@ unsafe fn get_hash_tbl_threaded_join_mut<T, H>(
 /// Probe the build table and add tuples to the results (inner join)
 fn probe_inner<T, F>(
     probe: &[T],
-    hash_tbls: &[HashMap<T, Vec<u32>, RandomState>],
+    hash_tbls: &[PlHashMap<T, Vec<u32>>],
     results: &mut Vec<(u32, u32)>,
     local_offset: usize,
     n_tables: u64,
@@ -104,11 +108,12 @@ fn probe_inner<T, F>(
     T: Send + Hash + Eq + Sync + Copy + AsU64,
     F: Fn(u32, u32) -> (u32, u32),
 {
+    assert!(hash_tbls.len().is_power_of_two());
     probe.iter().enumerate().for_each(|(idx_a, k)| {
         let idx_a = (idx_a + local_offset) as u32;
         // probe table that contains the hashed value
         let current_probe_table =
-            unsafe { get_hash_tbl_threaded_join(k.as_u64(), hash_tbls, n_tables) };
+            unsafe { get_hash_tbl_threaded_join_partitioned(k.as_u64(), hash_tbls, n_tables) };
 
         let value = current_probe_table.get(k);
 
@@ -119,26 +124,23 @@ fn probe_inner<T, F>(
     });
 }
 
-pub(crate) fn create_probe_table<T, IntoSlice>(
-    keys: Vec<IntoSlice>,
-) -> Vec<HashMap<T, Vec<u32>, RandomState>>
+pub(crate) fn create_probe_table<T, IntoSlice>(keys: Vec<IntoSlice>) -> Vec<PlHashMap<T, Vec<u32>>>
 where
     T: Send + Hash + Eq + Sync + Copy + AsU64,
     IntoSlice: AsRef<[T]> + Send + Sync,
 {
-    let n_threads = keys.len();
+    let n_partitions = set_partition_size();
 
     // We will create a hashtable in every thread.
     // We use the hash to partition the keys to the matching hashtable.
     // Every thread traverses all keys/hashes and ignores the ones that doesn't fall in that partition.
     POOL.install(|| {
-        (0..n_threads).into_par_iter().map(|thread_no| {
-            let thread_no = thread_no as u64;
+        (0..n_partitions).into_par_iter().map(|partition_no| {
+            let partition_no = partition_no as u64;
 
-            let mut hash_tbl: HashMap<T, Vec<u32>, RandomState> =
-                HashMap::with_capacity_and_hasher(HASHMAP_INIT_SIZE, Default::default());
+            let mut hash_tbl: PlHashMap<T, Vec<u32>> = PlHashMap::with_capacity(HASHMAP_INIT_SIZE);
 
-            let n_threads = n_threads as u64;
+            let n_partitions = n_partitions as u64;
             let mut offset = 0;
             for keys in &keys {
                 let keys = keys.as_ref();
@@ -149,7 +151,7 @@ where
                     let idx = cnt + offset;
                     cnt += 1;
 
-                    if this_thread(k.as_u64(), thread_no, n_threads) {
+                    if this_partition(k.as_u64(), partition_no, n_partitions) {
                         let entry = hash_tbl.entry(*k);
 
                         match entry {
@@ -187,6 +189,7 @@ where
     let hash_tbls = create_probe_table(build);
 
     let n_tables = hash_tbls.len() as u64;
+    debug_assert!(n_tables.is_power_of_two());
     let offsets = probe
         .iter()
         .map(|ph| ph.as_ref().len())
@@ -260,6 +263,7 @@ where
         .collect::<Vec<_>>();
 
     let n_tables = hash_tbls.len() as u64;
+    debug_assert!(n_tables.is_power_of_two());
 
     // next we probe the other relation
     POOL.install(|| {
@@ -279,8 +283,9 @@ where
                 probe.iter().enumerate().for_each(|(idx_a, k)| {
                     let idx_a = (idx_a + offset) as u32;
                     // probe table that contains the hashed value
-                    let current_probe_table =
-                        unsafe { get_hash_tbl_threaded_join(k.as_u64(), hash_tbls, n_tables) };
+                    let current_probe_table = unsafe {
+                        get_hash_tbl_threaded_join_partitioned(k.as_u64(), hash_tbls, n_tables)
+                    };
 
                     // we already hashed, so we don't have to hash again.
                     let value = current_probe_table.get(k);
@@ -304,7 +309,7 @@ where
 /// Probe the build table and add tuples to the results (inner join)
 fn probe_outer<T, F, G, H>(
     probe_hashes: &[Vec<(u64, T)>],
-    hash_tbls: &mut [HashMap<T, Vec<u32>, RandomState>],
+    hash_tbls: &mut [PlHashMap<T, Vec<u32>>],
     results: &mut Vec<(Option<u32>, Option<u32>)>,
     n_tables: u64,
     // Function that get index_a, index_b when there is a match and pushes to result
@@ -322,13 +327,15 @@ fn probe_outer<T, F, G, H>(
     // idx_b -> ...
     H: Fn(u32) -> (Option<u32>, Option<u32>),
 {
+    // needed for the partition shift instead of modulo to make sense
+    assert!(n_tables.is_power_of_two());
     let mut idx_a = 0;
     for probe_hashes in probe_hashes {
         for (h, key) in probe_hashes {
             let h = *h;
             // probe table that contains the hashed value
             let current_probe_table =
-                unsafe { get_hash_tbl_threaded_join_mut(h, hash_tbls, n_tables) };
+                unsafe { get_hash_tbl_threaded_join_mut_partitioned(h, hash_tbls, n_tables) };
 
             let entry = current_probe_table
                 .raw_entry_mut()
@@ -477,20 +484,13 @@ impl HashJoin<CategoricalType> for CategoricalChunked {
     }
 }
 
-fn n_join_threads() -> usize {
-    let max = std::env::var("POLARS_MAX_THREADS")
-        .map(|s| s.parse::<usize>().expect("integer"))
-        .unwrap_or(usize::MAX);
-    std::cmp::min(POOL.current_num_threads(), max)
-}
-
 fn num_group_join_inner<T>(left: &ChunkedArray<T>, right: &ChunkedArray<T>) -> Vec<(u32, u32)>
 where
     T: PolarsIntegerType,
     T::Native: Hash + Eq + Send + AsU64,
     Option<T::Native>: AsU64,
 {
-    let n_threads = n_join_threads();
+    let n_threads = POOL.current_num_threads();
     let (a, b, swap) = det_hash_prone_order!(left, right);
     let splitted_a = split_ca(a, n_threads).unwrap();
     let splitted_b = split_ca(b, n_threads).unwrap();
@@ -567,7 +567,7 @@ where
     T::Native: Hash + Eq + Send + AsU64,
     Option<T::Native>: AsU64,
 {
-    let n_threads = n_join_threads();
+    let n_threads = POOL.current_num_threads();
     let splitted_a = split_ca(left, n_threads).unwrap();
     let splitted_b = split_ca(right, n_threads).unwrap();
     match (
@@ -722,9 +722,9 @@ where
     fn hash_join_outer(&self, other: &ChunkedArray<T>) -> Vec<(Option<u32>, Option<u32>)> {
         let (a, b, swap) = det_hash_prone_order!(self, other);
 
-        let n_threads = n_join_threads();
-        let splitted_a = split_ca(a, n_threads).unwrap();
-        let splitted_b = split_ca(b, n_threads).unwrap();
+        let n_partitions = set_partition_size();
+        let splitted_a = split_ca(a, n_partitions).unwrap();
+        let splitted_b = split_ca(b, n_partitions).unwrap();
 
         match (a.null_count(), b.null_count()) {
             (0, 0) => {
@@ -763,9 +763,9 @@ impl HashJoin<BooleanType> for BooleanChunked {
     fn hash_join_outer(&self, other: &BooleanChunked) -> Vec<(Option<u32>, Option<u32>)> {
         let (a, b, swap) = det_hash_prone_order!(self, other);
 
-        let n_threads = n_join_threads();
-        let splitted_a = split_ca(a, n_threads).unwrap();
-        let splitted_b = split_ca(b, n_threads).unwrap();
+        let n_partitions = set_partition_size();
+        let splitted_a = split_ca(a, n_partitions).unwrap();
+        let splitted_b = split_ca(b, n_partitions).unwrap();
 
         match (a.null_count(), b.null_count()) {
             (0, 0) => {
@@ -836,9 +836,9 @@ impl HashJoin<Utf8Type> for Utf8Chunked {
     fn hash_join_outer(&self, other: &Utf8Chunked) -> Vec<(Option<u32>, Option<u32>)> {
         let (a, b, swap) = det_hash_prone_order!(self, other);
 
-        let n_threads = n_join_threads();
-        let splitted_a = split_ca(a, n_threads).unwrap();
-        let splitted_b = split_ca(b, n_threads).unwrap();
+        let n_partitions = set_partition_size();
+        let splitted_a = split_ca(a, n_partitions).unwrap();
+        let splitted_b = split_ca(b, n_partitions).unwrap();
 
         match (a.null_count(), b.null_count()) {
             (0, 0) => {
@@ -1233,6 +1233,7 @@ impl DataFrame {
 
 #[cfg(test)]
 mod test {
+    use crate::df;
     use crate::prelude::*;
     use crate::toggle_string_cache;
 
@@ -1312,12 +1313,28 @@ mod test {
     }
 
     #[test]
-    fn test_outer_join() {
+    fn test_outer_join() -> Result<()> {
         let (temp, rain) = create_frames();
-        let joined = temp.outer_join(&rain, "days", "days").unwrap();
+        let joined = temp.outer_join(&rain, "days", "days")?;
         println!("{:?}", &joined);
         assert_eq!(joined.height(), 5);
-        assert_eq!(joined.column("days").unwrap().sum::<i32>(), Some(7));
+        assert_eq!(joined.column("days")?.sum::<i32>(), Some(7));
+
+        let df_left = df!(
+                "a"=> ["a", "b", "a", "z"],
+                "b"=>[1, 2, 3, 4],
+                "c"=>[6, 5, 4, 3]
+        )?;
+        let df_right = df!(
+                "a"=> ["b", "c", "b", "a"],
+                "k"=> [0, 3, 9, 6],
+                "c"=> [1, 0, 2, 1]
+        )?;
+
+        let out = df_left.outer_join(&df_right, "a", "a")?;
+        assert_eq!(out.column("c_right")?.null_count(), 1);
+
+        Ok(())
     }
 
     #[test]
