@@ -21,13 +21,18 @@ mod arithmetic;
 pub mod explode;
 pub mod groupby;
 pub mod hash_join;
+#[cfg(feature = "rows")]
 pub mod row;
 pub mod select;
 mod upstream_traits;
+#[cfg(feature = "sort_multiple")]
 use crate::prelude::sort::prepare_argsort;
 use crate::POOL;
+#[cfg(feature = "serde")]
+use serde::{Deserialize, Serialize};
 
 #[derive(Clone)]
+#[cfg_attr(feature = "serde", derive(Serialize, Deserialize))]
 pub struct DataFrame {
     pub(crate) columns: Vec<Series>,
 }
@@ -130,6 +135,14 @@ impl DataFrame {
         DataFrame::new_no_checks(cols)
     }
 
+    /// Shrink the capacity of this DataFrame to fit it's length.
+    pub fn shrink_to_fit(&mut self) {
+        // Don't parallelize this. Memory overhead
+        for s in &mut self.columns {
+            s.shrink_to_fit();
+        }
+    }
+
     /// Aggregate all the chunks in the DataFrame to a single chunk.
     pub fn as_single_chunk(&mut self) -> &mut Self {
         // Don't parallelize this. Memory overhead
@@ -843,7 +856,7 @@ impl DataFrame {
     /// ```
     pub fn rename(&mut self, column: &str, name: &str) -> Result<&mut Self> {
         self.select_mut(column)
-            .ok_or_else(|| PolarsError::NotFound(name.to_string()))
+            .ok_or_else(|| PolarsError::NotFound(name.into()))
             .map(|s| s.rename(name))?;
         Ok(self)
     }
@@ -911,13 +924,13 @@ impl DataFrame {
         self.apply(column, |_| new_col.into_series())
     }
 
-    /// Replace or update a column.
+    /// Replace or update a column. The difference between this method and [DataFrame::with_column]
+    /// is that now the value of `column: &str` determines the name of the column and not the name
+    /// of the `Series` passed to this method.
     pub fn replace_or_add<S: IntoSeries>(&mut self, column: &str, new_col: S) -> Result<&mut Self> {
-        let new_col = new_col.into_series();
-        match self.replace(column, new_col.clone()) {
-            Err(_) => self.with_column(new_col),
-            Ok(_) => Ok(self),
-        }
+        let mut new_col = new_col.into_series();
+        new_col.rename(column);
+        self.with_column(new_col)
     }
 
     /// Replace column at index `idx` with a series.
@@ -1233,43 +1246,17 @@ impl DataFrame {
 
     /// Transform the underlying chunks in the DataFrame to Arrow RecordBatches
     pub fn as_record_batches(&self) -> Result<Vec<RecordBatch>> {
-        let n_chunks = self.n_chunks()?;
-        let width = self.width();
-
-        let schema = Arc::new(self.schema().to_arrow());
-
-        let mut record_batches = Vec::with_capacity(n_chunks);
-        for i in 0..n_chunks {
-            // the columns of a single recorbatch
-            let mut rb_cols = Vec::with_capacity(width);
-
-            for col in &self.columns {
-                rb_cols.push(Arc::clone(&col.chunks()[i]))
-            }
-            let rb = RecordBatch::try_new(Arc::clone(&schema), rb_cols)?;
-            record_batches.push(rb)
-        }
-        Ok(record_batches)
+        self.n_chunks()?;
+        Ok(self.iter_record_batches().collect())
     }
 
     /// Iterator over the rows in this DataFrame as Arrow RecordBatches.
-    pub fn iter_record_batches(
-        &mut self,
-        buffer_size: usize,
-    ) -> impl Iterator<Item = RecordBatch> + '_ {
-        match self.n_chunks() {
-            Ok(1) => {}
-            Ok(_) => {
-                self.columns = self.columns.iter().map(|s| s.rechunk()).collect();
-            }
-            Err(_) => {} // no data. So iterator will be empty
-        }
+    pub fn iter_record_batches(&self) -> impl Iterator<Item = RecordBatch> + '_ {
         RecordBatchIter {
             columns: &self.columns,
             schema: Arc::new(self.schema().to_arrow()),
-            buffer_size,
             idx: 0,
-            len: self.height(),
+            n_chunks: self.n_chunks().unwrap_or(0),
         }
     }
 
@@ -1365,6 +1352,8 @@ impl DataFrame {
     }
 
     /// Aggregate the column horizontally to their min values
+    #[cfg(feature = "zip_with")]
+    #[cfg_attr(docsrs, doc(cfg(feature = "zip_with")))]
     pub fn hmin(&self) -> Result<Option<Series>> {
         match self.columns.len() {
             0 => Ok(None),
@@ -1386,6 +1375,8 @@ impl DataFrame {
     }
 
     /// Aggregate the column horizontally to their max values
+    #[cfg(feature = "zip_with")]
+    #[cfg_attr(docsrs, doc(cfg(feature = "zip_with")))]
     pub fn hmax(&self) -> Result<Option<Series>> {
         match self.columns.len() {
             0 => Ok(None),
@@ -1636,35 +1627,27 @@ impl DataFrame {
 pub struct RecordBatchIter<'a> {
     columns: &'a Vec<Series>,
     schema: Arc<ArrowSchema>,
-    buffer_size: usize,
     idx: usize,
-    len: usize,
+    n_chunks: usize,
 }
 
 impl<'a> Iterator for RecordBatchIter<'a> {
     type Item = RecordBatch;
 
     fn next(&mut self) -> Option<Self::Item> {
-        if self.idx >= self.len {
-            return None;
-        }
-        // most iterations the slice length will be buffer_size, except for the last. That one
-        // may be shorter
-        let length = if self.idx + self.buffer_size < self.len {
-            self.buffer_size
+        if self.idx >= self.n_chunks {
+            None
         } else {
-            self.len - self.idx
-        };
+            // create a batch of the columns with the same chunk no.
+            let batch_cols = self
+                .columns
+                .iter()
+                .map(|s| s.chunks()[self.idx].clone())
+                .collect();
+            self.idx += 1;
 
-        let mut rb_cols = Vec::with_capacity(self.columns.len());
-        // take a slice from all columns and add the the current RecordBatch
-        self.columns.iter().for_each(|s| {
-            let slice = s.slice(self.idx as i64, length);
-            rb_cols.push(Arc::clone(&slice.chunks()[0]))
-        });
-        let rb = RecordBatch::try_new(Arc::clone(&self.schema), rb_cols).unwrap();
-        self.idx += length;
-        Some(rb)
+            Some(RecordBatch::try_new(self.schema.clone(), batch_cols).unwrap())
+        }
     }
 }
 
@@ -1781,14 +1764,12 @@ mod test {
 
     #[test]
     fn test_recordbatch_iterator() {
-        let mut df = df!(
+        let df = df!(
             "foo" => &[1, 2, 3, 4, 5]
         )
         .unwrap();
-        let mut iter = df.iter_record_batches(2);
-        assert_eq!(2, iter.next().unwrap().num_rows());
-        assert_eq!(2, iter.next().unwrap().num_rows());
-        assert_eq!(1, iter.next().unwrap().num_rows());
+        let mut iter = df.iter_record_batches();
+        assert_eq!(5, iter.next().unwrap().num_rows());
         assert!(iter.next().is_none());
     }
 
@@ -1937,6 +1918,7 @@ mod test {
     }
 
     #[test]
+    #[cfg(feature = "zip_with")]
     fn test_h_agg() {
         let a = Series::new("a", &[1, 2, 6]);
         let b = Series::new("b", &[Some(1), None, None]);
@@ -1960,4 +1942,18 @@ mod test {
             &[Some(4), Some(2), Some(6)]
         );
     }
+
+    #[test]
+    fn test_replace_or_add() -> Result<()> {
+        let mut df = df!(
+            "a" => [1, 2, 3],
+            "b" => [1, 2, 3]
+        )?;
+
+        // check that the new column is "c" and not "bar".
+        df.replace_or_add("c", Series::new("bar", [1, 2, 3]))?;
+
+        assert_eq!(df.get_column_names(), &["a", "b", "c"]);
+        Ok(())
+    }
 }
