@@ -14,7 +14,10 @@ use super::IntoSeries;
 use super::SeriesTrait;
 use crate::chunked_array::comparison::*;
 use crate::chunked_array::{
-    ops::aggregate::{ChunkAggSeries, VarAggSeries},
+    ops::{
+        aggregate::{ChunkAggSeries, VarAggSeries},
+        compare_inner::{IntoPartialEqInner, PartialEqInner},
+    },
     AsSinglePtr, ChunkIdIter,
 };
 use crate::fmt::FmtList;
@@ -23,6 +26,8 @@ use crate::frame::groupby::pivot::*;
 use crate::frame::groupby::*;
 use crate::frame::hash_join::{HashJoin, ZipOuterJoinColumn};
 use crate::prelude::*;
+#[cfg(feature = "checked_arithmetic")]
+use crate::series::arithmetic::checked::NumOpsDispatchChecked;
 use ahash::RandomState;
 use arrow::array::{ArrayData, ArrayRef};
 use arrow::buffer::Buffer;
@@ -64,15 +69,23 @@ macro_rules! impl_dyn_series {
                 self.0.equal_element(idx_self, idx_other, other)
             }
 
+            #[cfg(feature = "zip_with")]
             fn zip_with_same_type(&self, mask: &BooleanChunked, other: &Series) -> Result<Series> {
                 ChunkZip::zip_with(&self.0, mask, other.as_ref().as_ref())
                     .map(|ca| ca.into_series())
             }
+            fn into_partial_eq_inner<'a>(&'a self) -> Box<dyn PartialEqInner + 'a> {
+                (&self.0).into_partial_eq_inner()
+            }
 
-            fn vec_hash(&self, random_state: RandomState) -> UInt64Chunked {
+            fn vec_hash(&self, random_state: RandomState) -> AlignedVec<u64> {
                 self.0.vec_hash(random_state)
             }
 
+            fn vec_hash_combine(&self, build_hasher: RandomState, hashes: &mut [u64]) {
+                self.0.vec_hash_combine(build_hasher, hashes)
+            }
+
             fn agg_mean(&self, groups: &[(u32, Vec<u32>)]) -> Option<Series> {
                 self.0.agg_mean(groups)
             }
@@ -128,7 +141,7 @@ macro_rules! impl_dyn_series {
             #[cfg(feature = "pivot")]
             fn pivot<'a>(
                 &self,
-                pivot_series: &'a (dyn SeriesTrait + 'a),
+                pivot_series: &'a Series,
                 keys: Vec<Series>,
                 groups: &[(u32, Vec<u32>)],
                 agg_type: PivotAgg,
@@ -139,7 +152,7 @@ macro_rules! impl_dyn_series {
             #[cfg(feature = "pivot")]
             fn pivot_count<'a>(
                 &self,
-                pivot_series: &'a (dyn SeriesTrait + 'a),
+                pivot_series: &'a Series,
                 keys: Vec<Series>,
                 groups: &[(u32, Vec<u32>)],
             ) -> Result<DataFrame> {
@@ -226,6 +239,9 @@ macro_rules! impl_dyn_series {
             fn chunks(&self) -> &Vec<ArrayRef> {
                 self.0.chunks()
             }
+            fn shrink_to_fit(&mut self) {
+                self.0.shrink_to_fit()
+            }
 
             fn i8(&self) -> Result<&Int8Chunked> {
                 if matches!(self.0.dtype(), DataType::Int8) {
@@ -690,7 +706,7 @@ macro_rules! impl_dyn_series {
                 ChunkUnique::is_duplicated(&self.0)
             }
 
-            fn null_bits(&self) -> Vec<(usize, Option<Buffer>)> {
+            fn null_bits(&self) -> Vec<(usize, Option<&Buffer>)> {
                 self.0.null_bits().collect()
             }
 
@@ -824,15 +840,21 @@ macro_rules! impl_dyn_series {
             fn is_in(&self, other: &Series) -> Result<BooleanChunked> {
                 IsIn::is_in(&self.0, other)
             }
+            #[cfg(feature = "repeat_by")]
+            fn repeat_by(&self, by: &UInt32Chunked) -> ListChunked {
+                RepeatBy::repeat_by(&self.0, by)
+            }
+
+            #[cfg(feature = "checked_arithmetic")]
+            fn checked_div(&self, rhs: &Series) -> Result<Series> {
+                self.0.checked_div(rhs)
+            }
         }
     };
 }
 
 impl_dyn_series!(Float32Chunked);
 impl_dyn_series!(Float64Chunked);
-impl_dyn_series!(Utf8Chunked);
-impl_dyn_series!(ListChunked);
-impl_dyn_series!(BooleanChunked);
 #[cfg(feature = "dtype-u8")]
 impl_dyn_series!(UInt8Chunked);
 #[cfg(feature = "dtype-u16")]
@@ -847,3 +869,43 @@ impl_dyn_series!(Int16Chunked);
 impl_dyn_series!(Int32Chunked);
 impl_dyn_series!(Int64Chunked);
 impl_dyn_series!(CategoricalChunked);
+impl_dyn_series!(Utf8Chunked);
+impl_dyn_series!(ListChunked);
+impl_dyn_series!(BooleanChunked);
+
+macro_rules! impl_dyn_series_numeric {
+    ($ca: ident) => {
+        impl private::PrivateSeriesNumeric for SeriesWrap<$ca> {
+            fn bit_repr_is_large(&self) -> bool {
+                $ca::bit_repr_is_large()
+            }
+            fn bit_repr_large(&self) -> UInt64Chunked {
+                self.0.bit_repr_large()
+            }
+            fn bit_repr_small(&self) -> UInt32Chunked {
+                self.0.bit_repr_small()
+            }
+        }
+    };
+}
+
+impl_dyn_series_numeric!(Float32Chunked);
+impl_dyn_series_numeric!(Float64Chunked);
+#[cfg(feature = "dtype-u8")]
+impl_dyn_series_numeric!(UInt8Chunked);
+#[cfg(feature = "dtype-u16")]
+impl_dyn_series_numeric!(UInt16Chunked);
+impl_dyn_series_numeric!(UInt32Chunked);
+#[cfg(feature = "dtype-u64")]
+impl_dyn_series_numeric!(UInt64Chunked);
+#[cfg(feature = "dtype-i8")]
+impl_dyn_series_numeric!(Int8Chunked);
+#[cfg(feature = "dtype-i16")]
+impl_dyn_series_numeric!(Int16Chunked);
+impl_dyn_series_numeric!(Int32Chunked);
+impl_dyn_series_numeric!(Int64Chunked);
+impl_dyn_series_numeric!(CategoricalChunked);
+
+impl private::PrivateSeriesNumeric for SeriesWrap<BooleanChunked> {}
+impl private::PrivateSeriesNumeric for SeriesWrap<Utf8Chunked> {}
+impl private::PrivateSeriesNumeric for SeriesWrap<ListChunked> {}
