@@ -7,7 +7,7 @@ mod comparison;
 pub mod implementations;
 pub(crate) mod iterator;
 
-use crate::chunked_array::{builder::get_list_builder, float::IsNan, ChunkIdIter};
+use crate::chunked_array::{builder::get_list_builder, ChunkIdIter};
 use crate::utils::{split_ca, split_series};
 use crate::{series::arithmetic::coerce_lhs_rhs, POOL};
 use arrow::array::ArrayData;
@@ -32,9 +32,22 @@ pub(crate) mod private {
     use crate::frame::groupby::pivot::PivotAgg;
     use crate::frame::groupby::GroupTuples;
 
+    use crate::chunked_array::ops::compare_inner::PartialEqInner;
     use ahash::RandomState;
     use std::borrow::Cow;
 
+    pub trait PrivateSeriesNumeric {
+        fn bit_repr_is_large(&self) -> bool {
+            unimplemented!()
+        }
+        fn bit_repr_large(&self) -> UInt64Chunked {
+            unimplemented!()
+        }
+        fn bit_repr_small(&self) -> UInt32Chunked {
+            unimplemented!()
+        }
+    }
+
     pub trait PrivateSeries {
         unsafe fn equal_element(
             &self,
@@ -44,7 +57,14 @@ pub(crate) mod private {
         ) -> bool {
             unimplemented!()
         }
-        fn vec_hash(&self, _build_hasher: RandomState) -> UInt64Chunked {
+        #[allow(clippy::wrong_self_convention)]
+        fn into_partial_eq_inner<'a>(&'a self) -> Box<dyn PartialEqInner + 'a> {
+            unimplemented!()
+        }
+        fn vec_hash(&self, _build_hasher: RandomState) -> AlignedVec<u64> {
+            unimplemented!()
+        }
+        fn vec_hash_combine(&self, _build_hasher: RandomState, _hashes: &mut [u64]) {
             unimplemented!()
         }
         fn agg_mean(&self, _groups: &[(u32, Vec<u32>)]) -> Option<Series> {
@@ -90,7 +110,7 @@ pub(crate) mod private {
         #[cfg(feature = "pivot")]
         fn pivot<'a>(
             &self,
-            _pivot_series: &'a (dyn SeriesTrait + 'a),
+            _pivot_series: &'a Series,
             _keys: Vec<Series>,
             _groups: &[(u32, Vec<u32>)],
             _agg_type: PivotAgg,
@@ -101,7 +121,7 @@ pub(crate) mod private {
         #[cfg(feature = "pivot")]
         fn pivot_count<'a>(
             &self,
-            _pivot_series: &'a (dyn SeriesTrait + 'a),
+            _pivot_series: &'a Series,
             _keys: Vec<Series>,
             _groups: &[(u32, Vec<u32>)],
         ) -> Result<DataFrame> {
@@ -159,7 +179,9 @@ pub(crate) mod private {
     }
 }
 
-pub trait SeriesTrait: Send + Sync + private::PrivateSeries {
+pub trait SeriesTrait:
+    Send + Sync + private::PrivateSeries + private::PrivateSeriesNumeric
+{
     /// Get an array with the cumulative max computed at every element
     fn cum_max(&self, _reverse: bool) -> Series {
         panic!("operation cum_max not supported for this dtype")
@@ -212,6 +234,14 @@ pub trait SeriesTrait: Send + Sync + private::PrivateSeries {
         self.chunks().len()
     }
 
+    /// Shrink the capacity of this array to fit it's length.
+    fn shrink_to_fit(&mut self) {
+        eprintln!(
+            "shrink to fit, not yet supported for this {:?}",
+            self.dtype()
+        )
+    }
+
     /// Unpack to ChunkedArray of dtype i8
     fn i8(&self) -> Result<&Int8Chunked> {
         Err(PolarsError::DataTypeMisMatch(
@@ -358,6 +388,40 @@ pub trait SeriesTrait: Send + Sync + private::PrivateSeries {
         ))
     }
 
+    /// Check if underlying physical data is numeric.
+    ///
+    /// Date types and Categoricals are also considered numeric.
+    fn is_numeric_physical(&self) -> bool {
+        // allow because it cannot be replaced when object feature is activated
+        #[allow(clippy::match_like_matches_macro)]
+        match self.dtype() {
+            DataType::Utf8 | DataType::List(_) | DataType::Boolean | DataType::Null => false,
+            #[cfg(feature = "object")]
+            DataType::Object(_) => false,
+            _ => true,
+        }
+    }
+
+    /// Check if underlying data is numeric
+    fn is_numeric(&self) -> bool {
+        // allow because it cannot be replaced when object feature is activated
+        #[allow(clippy::match_like_matches_macro)]
+        match self.dtype() {
+            DataType::Utf8
+            | DataType::List(_)
+            | DataType::Categorical
+            | DataType::Date32
+            | DataType::Date64
+            | DataType::Duration(_)
+            | DataType::Time64(_)
+            | DataType::Boolean
+            | DataType::Null => false,
+            #[cfg(feature = "object")]
+            DataType::Object(_) => false,
+            _ => true,
+        }
+    }
+
     /// Append Arrow array of same dtype to this Series.
     fn append_array(&mut self, _other: ArrayRef) -> Result<()> {
         unimplemented!()
@@ -605,7 +669,7 @@ pub trait SeriesTrait: Send + Sync + private::PrivateSeries {
     }
 
     /// Get the bits that represent the null values of the underlying ChunkedArray
-    fn null_bits(&self) -> Vec<(usize, Option<Buffer>)> {
+    fn null_bits(&self) -> Vec<(usize, Option<&Buffer>)> {
         unimplemented!()
     }
 
@@ -963,6 +1027,16 @@ pub trait SeriesTrait: Send + Sync + private::PrivateSeries {
     fn is_in(&self, _other: &Series) -> Result<BooleanChunked> {
         unimplemented!()
     }
+    #[cfg(feature = "repeat_by")]
+    #[cfg_attr(docsrs, doc(cfg(feature = "repeat_by")))]
+    fn repeat_by(&self, _by: &UInt32Chunked) -> ListChunked {
+        unimplemented!()
+    }
+    #[cfg(feature = "checked_arithmetic")]
+    #[cfg_attr(docsrs, doc(cfg(feature = "checked_arithmetic")))]
+    fn checked_div(&self, _rhs: &Series) -> Result<Series> {
+        unimplemented!()
+    }
 }
 
 impl<'a> (dyn SeriesTrait + 'a) {
@@ -981,7 +1055,11 @@ impl<'a> (dyn SeriesTrait + 'a) {
 }
 
 /// # Series
-/// The columnar data type for a DataFrame. The [Series enum](enum.Series.html) consists
+/// The columnar data type for a DataFrame.
+///
+/// Most of the available functions are definedin the [SeriesTrait trait](crate::series::SeriesTrait).
+///
+/// The `Series` struct consists
 /// of typed [ChunkedArray](../chunked_array/struct.ChunkedArray.html)'s. To quickly cast
 /// a `Series` to a `ChunkedArray` you can call the method with the name of the type:
 ///
@@ -1090,6 +1168,11 @@ impl Series {
         self
     }
 
+    /// Shrink the capacity of this array to fit it's length.
+    pub fn shrink_to_fit(&mut self) {
+        self.get_inner_mut().shrink_to_fit()
+    }
+
     /// Append arrow array of same datatype.
     pub fn append_array(&mut self, other: ArrayRef) -> Result<&mut Self> {
         self.get_inner_mut().append_array(other)?;
@@ -1248,6 +1331,8 @@ impl Series {
 
     /// Create a new ChunkedArray with values from self where the mask evaluates `true` and values
     /// from `other` where the mask evaluates `false`
+    #[cfg(feature = "zip_with")]
+    #[cfg_attr(docsrs, doc(cfg(feature = "zip_with")))]
     pub fn zip_with(&self, mask: &BooleanChunked, other: &Series) -> Result<Series> {
         let (lhs, rhs) = coerce_lhs_rhs(self, other)?;
         lhs.zip_with_same_type(mask, rhs.as_ref())
@@ -1282,7 +1367,7 @@ impl Series {
         idx: &UInt32Chunked,
         rechunk: bool,
     ) -> Result<Series> {
-        let n_threads = num_cpus::get();
+        let n_threads = POOL.current_num_threads();
         let idx = split_ca(idx, n_threads)?;
 
         let series: Result<Vec<_>> =
@@ -1308,7 +1393,7 @@ impl Series {
     ///
     /// Out of bounds access doesn't Error but will return a Null value
     pub fn take_threaded(&self, idx: &UInt32Chunked, rechunk: bool) -> Series {
-        let n_threads = num_cpus::get();
+        let n_threads = POOL.current_num_threads();
         let idx = split_ca(idx, n_threads).unwrap();
 
         let series: Vec<_> = POOL.install(|| idx.par_iter().map(|idx| self.take(idx)).collect());
@@ -1329,7 +1414,13 @@ impl Series {
 
     /// Filter by boolean mask. This operation clones data.
     pub fn filter_threaded(&self, filter: &BooleanChunked, rechunk: bool) -> Result<Series> {
-        let n_threads = num_cpus::get();
+        // this would fail if there is a broadcasting filter.
+        // because we cannot split that filter over threads
+        // besides they are a no-op, so we do the standard filter.
+        if filter.len() == 1 {
+            return self.filter(filter);
+        }
+        let n_threads = POOL.current_num_threads();
         let filters = split_ca(filter, n_threads).unwrap();
         let series = split_series(self, n_threads).unwrap();
 
@@ -1354,6 +1445,30 @@ impl Series {
             Ok(s)
         }
     }
+
+    /// Round underlying floating point array to given decimal.
+    #[cfg(feature = "round_series")]
+    #[cfg_attr(docsrs, doc(cfg(feature = "round_series")))]
+    pub fn round(&self, decimals: u32) -> Result<Self> {
+        use num::traits::Pow;
+        if let Ok(ca) = self.f32() {
+            let multiplier = 10.0.pow(decimals as f32) as f32;
+            let s = ca
+                .apply(|val| (val * multiplier).round() / multiplier)
+                .into_series();
+            return Ok(s);
+        }
+        if let Ok(ca) = self.f64() {
+            let multiplier = 10.0.pow(decimals as f32) as f64;
+            let s = ca
+                .apply(|val| (val * multiplier).round() / multiplier)
+                .into_series();
+            return Ok(s);
+        }
+        Err(PolarsError::DataTypeMisMatch(
+            format!("{:?} is not a floating point datatype", self.dtype()).into(),
+        ))
+    }
 }
 
 impl Deref for Series {
@@ -1477,21 +1592,11 @@ impl std::convert::TryFrom<(&str, Vec<ArrayRef>)> for Series {
                     .collect_vec();
                 Ok(Utf8Chunked::new_from_chunks(name, chunks).into_series())
             }
-            ArrowDataType::List(_) => {
+            ArrowDataType::List(fld) => {
                 let chunks = chunks
                     .iter()
-                    .map(|arr| {
-                        cast(
-                            arr,
-                            &ArrowDataType::LargeList(Box::new(arrow::datatypes::Field::new(
-                                "",
-                                ArrowDataType::Null,
-                                true,
-                            ))),
-                        )
-                        .unwrap()
-                    })
-                    .collect_vec();
+                    .map(|arr| cast(arr, &ArrowDataType::LargeList(fld.clone())).unwrap())
+                    .collect();
                 Ok(ListChunked::new_from_chunks(name, chunks).into_series())
             }
             ArrowDataType::Boolean => {
@@ -1543,6 +1648,14 @@ impl std::convert::TryFrom<(&str, Vec<ArrayRef>)> for Series {
                 #[cfg(not(feature = "dtype-i8"))]
                 Ok(UInt32Chunked::full_null(name, len).into_series())
             }
+            #[cfg(feature = "dtype-date64")]
+            ArrowDataType::Timestamp(TimeUnit::Millisecond, None) => {
+                let chunks = chunks
+                    .iter()
+                    .map(|arr| cast(arr, &ArrowDataType::Date64).unwrap())
+                    .collect();
+                Ok(Date64Chunked::new_from_chunks(name, chunks).into_series())
+            }
             dt => Err(PolarsError::InvalidOperation(
                 format!("Cannot create polars series from {:?} type", dt).into(),
             )),
@@ -1673,4 +1786,13 @@ mod test {
         series.slice(-6, 2);
         series.slice(4, 2);
     }
+
+    #[test]
+    #[cfg(feature = "round_series")]
+    fn test_round_series() {
+        let series = Series::new("a", &[1.003, 2.23222, 3.4352]);
+        let out = series.round(2).unwrap();
+        let ca = out.f64().unwrap();
+        assert_eq!(ca.get(0), Some(1.0));
+    }
 }
