@@ -5,6 +5,7 @@ pub use arrow;
 pub use chrono;
 pub use num_cpus;
 pub use polars_arrow::utils::TrustMyLength;
+pub use polars_arrow::utils::*;
 pub use rayon;
 use rayon::prelude::*;
 use std::borrow::Cow;
@@ -35,40 +36,17 @@ fn index_of<T>(slice: &[T], item: &T) -> Option<usize> {
     }
 }
 
-/// Used to split the mantissa and exponent of floating point numbers
-/// https://stackoverflow.com/questions/39638363/how-can-i-use-a-hashmap-with-f64-as-key-in-rust
-pub(crate) fn integer_decode_f64(val: f64) -> (u64, i16, i8) {
-    let bits: u64 = val.to_bits();
-    let sign: i8 = if bits >> 63 == 0 { 1 } else { -1 };
-    let mut exponent: i16 = ((bits >> 52) & 0x7ff) as i16;
-    let mantissa = if exponent == 0 {
-        (bits & 0xfffffffffffff) << 1
-    } else {
-        (bits & 0xfffffffffffff) | 0x10000000000000
-    };
-
-    exponent -= 1023 + 52;
-    (mantissa, exponent, sign)
-}
-
-/// Returns the mantissa, exponent and sign as integers.
-/// https://github.com/rust-lang/rust/blob/5c674a11471ec0569f616854d715941757a48a0a/src/libcore/num/f32.rs#L203-L216
-pub(crate) fn integer_decode_f32(val: f32) -> (u64, i16, i8) {
-    let bits: u32 = val.to_bits();
-    let sign: i8 = if bits >> 31 == 0 { 1 } else { -1 };
-    let mut exponent: i16 = ((bits >> 23) & 0xff) as i16;
-    let mantissa = if exponent == 0 {
-        (bits & 0x7fffff) << 1
-    } else {
-        (bits & 0x7fffff) | 0x800000
-    };
-    // Exponent bias + mantissa shift
-    exponent -= 127 + 23;
-    (mantissa as u64, exponent, sign)
-}
-
-pub(crate) fn floating_encode_f64(mantissa: u64, exponent: i16, sign: i8) -> f64 {
-    sign as f64 * mantissa as f64 * (2.0f64).powf(exponent as f64)
+pub(crate) fn set_partition_size() -> usize {
+    let mut n_partitions = POOL.current_num_threads();
+    // set n_partitions to closes 2^n above the no of threads.
+    loop {
+        if n_partitions.is_power_of_two() {
+            break;
+        } else {
+            n_partitions += 1;
+        }
+    }
+    n_partitions
 }
 
 /// Just a wrapper structure. Useful for certain impl specializations
@@ -104,7 +82,7 @@ impl<T> DerefMut for NoNull<T> {
     }
 }
 
-pub fn get_iter_capacity<T, I: Iterator<Item = T>>(iter: &I) -> usize {
+pub(crate) fn get_iter_capacity<T, I: Iterator<Item = T>>(iter: &I) -> usize {
     match iter.size_hint() {
         (_lower, Some(upper)) => upper,
         (0, None) => 1024,
@@ -135,14 +113,17 @@ macro_rules! split_array {
     }};
 }
 
+#[cfg(feature = "private")]
 pub fn split_ca<T>(ca: &ChunkedArray<T>, n: usize) -> Result<Vec<ChunkedArray<T>>> {
     split_array!(ca, n, i64)
 }
 
+#[cfg(feature = "private")]
 pub fn split_series(s: &Series, n: usize) -> Result<Vec<Series>> {
     split_array!(s, n, i64)
 }
 
+#[cfg(feature = "private")]
 pub fn split_df(df: &DataFrame, n: usize) -> Result<Vec<DataFrame>> {
     trait Len {
         fn len(&self) -> usize;
@@ -156,6 +137,7 @@ pub fn split_df(df: &DataFrame, n: usize) -> Result<Vec<DataFrame>> {
 }
 
 #[inline]
+#[cfg(feature = "private")]
 pub fn slice_offsets(offset: i64, length: usize, array_len: usize) -> (usize, usize) {
     let abs_offset = offset.abs() as usize;
 
@@ -187,6 +169,7 @@ impl Default for Node {
 }
 
 #[derive(Clone)]
+#[cfg(feature = "private")]
 pub struct Arena<T> {
     items: Vec<T>,
 }
@@ -275,47 +258,7 @@ impl<T: Default> Arena<T> {
     }
 }
 
-/// An iterator that iterates an unknown at compile time number
-/// of iterators simultaneously.
-///
-/// IMPORTANT: It differs from `std::iter::Zip` in the return type
-/// of `next`. It returns a `Vec` instead of a `tuple`, which implies
-/// that the result is non-copiable anymore.
-pub struct DynamicZip<I>
-where
-    I: Iterator,
-{
-    iterators: Vec<I>,
-}
-
-impl<I, T> Iterator for DynamicZip<I>
-where
-    I: Iterator<Item = T>,
-{
-    type Item = Vec<T>;
-
-    fn next(&mut self) -> Option<Self::Item> {
-        self.iterators.iter_mut().map(|iter| iter.next()).collect()
-    }
-}
-
-/// A trait to convert a value to a `DynamicZip`.
-pub trait IntoDynamicZip<I>
-where
-    I: Iterator,
-{
-    fn into_dynamic_zip(self) -> DynamicZip<I>;
-}
-
-impl<I> IntoDynamicZip<I> for Vec<I>
-where
-    I: Iterator,
-{
-    fn into_dynamic_zip(self) -> DynamicZip<I> {
-        DynamicZip { iterators: self }
-    }
-}
-
+/// Apply a macro on the Series
 #[macro_export]
 macro_rules! match_arrow_data_type_apply_macro {
     ($obj:expr, $macro:ident, $macro_utf8:ident, $macro_bool:ident $(, $opt_args:expr)*) => {{
@@ -337,9 +280,7 @@ macro_rules! match_arrow_data_type_apply_macro {
             DataType::Int64 => $macro!(Int64Type $(, $opt_args)*),
             DataType::Float32 => $macro!(Float32Type $(, $opt_args)*),
             DataType::Float64 => $macro!(Float64Type $(, $opt_args)*),
-            #[cfg(feature = "dtype-date32")]
             DataType::Date32 => $macro!(Date32Type $(, $opt_args)*),
-            #[cfg(feature = "dtype-date64")]
             DataType::Date64 => $macro!(Date64Type $(, $opt_args)*),
             #[cfg(feature = "dtype-time64-ns")]
             DataType::Time64(TimeUnit::Nanosecond) => $macro!(Time64NanosecondType $(, $opt_args)*),
@@ -352,6 +293,7 @@ macro_rules! match_arrow_data_type_apply_macro {
     }};
 }
 
+/// Apply a macro on the Downcasted ChunkedArray's
 #[macro_export]
 macro_rules! match_arrow_data_type_apply_macro_ca {
     ($self:expr, $macro:ident, $macro_utf8:ident, $macro_bool:ident $(, $opt_args:expr)*) => {{
@@ -388,6 +330,32 @@ macro_rules! match_arrow_data_type_apply_macro_ca {
     }};
 }
 
+/// Apply a macro on the Downcasted ChunkedArray's of DataTypes that are logical numerics.
+/// So no dates.
+#[macro_export]
+macro_rules! match_arrow_data_type_apply_macro_ca_logical_num {
+    ($self:expr, $macro:ident $(, $opt_args:expr)*) => {{
+        match $self.dtype() {
+            #[cfg(feature = "dtype-u8")]
+            DataType::UInt8 => $macro!($self.u8().unwrap() $(, $opt_args)*),
+            #[cfg(feature = "dtype-u16")]
+            DataType::UInt16 => $macro!($self.u16().unwrap() $(, $opt_args)*),
+            DataType::UInt32 => $macro!($self.u32().unwrap() $(, $opt_args)*),
+            #[cfg(feature = "dtype-u64")]
+            DataType::UInt64 => $macro!($self.u64().unwrap() $(, $opt_args)*),
+            #[cfg(feature = "dtype-i8")]
+            DataType::Int8 => $macro!($self.i8().unwrap() $(, $opt_args)*),
+            #[cfg(feature = "dtype-i16")]
+            DataType::Int16 => $macro!($self.i16().unwrap() $(, $opt_args)*),
+            DataType::Int32 => $macro!($self.i32().unwrap() $(, $opt_args)*),
+            DataType::Int64 => $macro!($self.i64().unwrap() $(, $opt_args)*),
+            DataType::Float32 => $macro!($self.f32().unwrap() $(, $opt_args)*),
+            DataType::Float64 => $macro!($self.f64().unwrap() $(, $opt_args)*),
+            _ => unimplemented!(),
+        }
+    }};
+}
+
 #[macro_export]
 macro_rules! apply_method_all_arrow_series {
     ($self:expr, $method:ident, $($args:expr),*) => {
@@ -409,9 +377,7 @@ macro_rules! apply_method_all_arrow_series {
             DataType::Int64 => $self.i64().unwrap().$method($($args),*),
             DataType::Float32 => $self.f32().unwrap().$method($($args),*),
             DataType::Float64 => $self.f64().unwrap().$method($($args),*),
-            #[cfg(feature = "dtype-date32")]
             DataType::Date32 => $self.date32().unwrap().$method($($args),*),
-            #[cfg(feature = "dtype-date64")]
             DataType::Date64 => $self.date64().unwrap().$method($($args),*),
             #[cfg(feature = "dtype-time64-ns")]
             DataType::Time64(TimeUnit::Nanosecond) => $self.time64_nanosecond().unwrap().$method($($args),*),
@@ -420,7 +386,7 @@ macro_rules! apply_method_all_arrow_series {
             #[cfg(feature = "dtype-duration-ms")]
             DataType::Duration(TimeUnit::Millisecond) => $self.duration_millisecond().unwrap().$method($($args),*),
             DataType::List(_) => $self.list().unwrap().$method($($args),*),
-            _ => unimplemented!()
+            dt => panic!("dtype {:?} not supported", dt)
         }
     }
 }
@@ -516,6 +482,7 @@ macro_rules! df {
 }
 
 /// Given two datatypes, determine the supertype that both types can safely be cast to
+#[cfg(feature = "private")]
 pub fn get_supertype(l: &DataType, r: &DataType) -> Result<DataType> {
     match _get_supertype(l, r) {
         Some(dt) => Ok(dt),
@@ -719,14 +686,9 @@ pub fn accumulate_dataframes_horizontal(dfs: Vec<DataFrame>) -> Result<DataFrame
     Ok(acc_df)
 }
 
-#[cfg(target_os = "linux")]
-extern "C" {
-    #[allow(dead_code)]
-    pub fn malloc_trim(__pad: usize) -> std::os::raw::c_int;
-}
-
 /// Simple wrapper to parallelize functions that can be divided over threads aggregated and
 /// finally aggregated in the main thread. This can be done for sum, min, max, etc.
+#[cfg(feature = "private")]
 pub fn parallel_op_series<F>(f: F, s: Series, n_threads: Option<usize>) -> Result<Series>
 where
     F: Fn(Series) -> Result<Series> + Send + Sync,
@@ -746,26 +708,6 @@ where
     f(out)
 }
 
-pub trait CustomIterTools: Iterator {
-    fn fold_first_<F>(mut self, f: F) -> Option<Self::Item>
-    where
-        Self: Sized,
-        F: FnMut(Self::Item, Self::Item) -> Self::Item,
-    {
-        let first = self.next()?;
-        Some(self.fold(first, f))
-    }
-
-    fn trust_my_length(self, length: usize) -> TrustMyLength<Self, Self::Item>
-    where
-        Self: Sized,
-    {
-        TrustMyLength::new(self, length)
-    }
-}
-
-impl<T: ?Sized> CustomIterTools for T where T: Iterator {}
-
 pub(crate) fn align_chunks_binary<'a, T, B>(
     left: &'a ChunkedArray<T>,
     right: &'a ChunkedArray<B>,
@@ -887,22 +829,25 @@ impl<T> IntoVec<T> for Vec<T> {
 
 /// This logic is same as the impl on ChunkedArray
 /// The difference is that there is less indirection because the caller should preallocate
-/// `chunk_lens` once. On the `ChunkedArray` we indicrect through an `ArrayRef` which is an indirection
+/// `chunk_lens` once. On the `ChunkedArray` we indirect through an `ArrayRef` which is an indirection
 /// and a vtable.
 #[inline]
-pub(crate) fn index_to_chunked_index<I: Iterator<Item = usize>>(
+pub(crate) fn index_to_chunked_index<
+    I: Iterator<Item = Idx>,
+    Idx: PartialOrd + std::ops::AddAssign + std::ops::SubAssign + num::Zero + num::One,
+>(
     chunk_lens: I,
-    index: usize,
-) -> (usize, usize) {
+    index: Idx,
+) -> (Idx, Idx) {
     let mut index_remainder = index;
-    let mut current_chunk_idx = 0;
+    let mut current_chunk_idx = num::Zero::zero();
 
     for chunk_len in chunk_lens {
         if chunk_len > index_remainder {
             break;
         } else {
             index_remainder -= chunk_len;
-            current_chunk_idx += 1;
+            current_chunk_idx += num::One::one();
         }
     }
     (current_chunk_idx, index_remainder)
